##
# File: OeCommonUtils.py
# Date: 22-Oct-2020
#
##
import logging

from openeye import oechem

logger = logging.getLogger(__name__)


class OeCommonUtils(object):
    @staticmethod
    def getAtomBondExprOpts(matchOpts):
        """Return the OE constants for atom and bond matching criteria.

        Args:
            matchOpts (string): qualitative description of atom and bond matching criteria

        Returns:
            tuple: OE atom and bond comparison expressions
        """
        if matchOpts in ["default", "strict", "graph-strict", "graph-default", "sub-struct-graph-strict"]:
            atomexpr = oechem.OEExprOpts_AtomicNumber | oechem.OEExprOpts_FormalCharge | oechem.OEExprOpts_Chiral | oechem.OEExprOpts_Aromaticity
            bondexpr = oechem.OEExprOpts_BondOrder | oechem.OEExprOpts_Aromaticity | oechem.OEExprOpts_Chiral
        elif matchOpts in ["relaxed-stereo", "graph-relaxed-stereo", "sub-struct-graph-relaxed-stereo"]:
            atomexpr = oechem.OEExprOpts_AtomicNumber | oechem.OEExprOpts_Chiral | oechem.OEExprOpts_FormalCharge
            bondexpr = oechem.OEExprOpts_BondOrder | oechem.OEExprOpts_Chiral
        elif matchOpts in ["relaxed", "graph-relaxed", "simple", "sub-struct-graph-relaxed"]:
            atomexpr = oechem.OEExprOpts_AtomicNumber | oechem.OEExprOpts_FormalCharge
            bondexpr = oechem.OEExprOpts_BondOrder
        elif matchOpts in ["exact"]:
            atomexpr = oechem.OEExprOpts_ExactAtoms
            bondexpr = oechem.OEExprOpts_ExactBonds
        else:
            logger.error("Unanticipated match options %r", matchOpts)
            atomexpr = oechem.OEExprOpts_DefaultAtoms
            bondexpr = oechem.OEExprOpts_DefaultBonds
            logger.error("Unanticipated match options %r", matchOpts)
        #
        #
        # oechem.OEExprOpts_EqSingleDouble
        # atomexpr = oechem.OEExprOpts_AtomicNumber|oechem.OEExprOpts_EqAromatic
        # bondexpr = 0
        #
        # atomexpr = oechem.OEExprOpts_AtomicNumber|oechem.OEExprOpts_Aromaticity
        # bondexpr = oechem.OEExprOpts_BondOrder|oechem.OEExprOpts_EqNotAromatic
        #
        return atomexpr, bondexpr



## File: cactvsAnnotateMol.py
# Date: Jan 7, 2020 jdw
#
# $cactvsPath/cspy cactvsAnnotateMol.py <input sdf path> <output json path>
#
# This module is using an external self-contained python library in CACTVS bundle
##
# flake8: noqa
# pylint: skip-file
import json
import logging
import os
import sys

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


def annotateCactvs(molFilePath, aroModel="cactvs"):
    retD = {}
    try:
        cactvs["aromaticity_model"] = aroModel if aroModel in ["cactvs", "daylight", "tripos"] else "cactvs"
        logger.info("aroModel is %r", aroModel)
        #
        eh = Molfile(molFilePath).read()
        # print(type(eh))
        # print(eh.atoms())
        atomL = []
        iChargeTotal = 0
        for ii, at in enumerate(eh.atoms(), 1):
            aD = {}
            st = at.A_LABEL_STEREO != "undef"
            aD = {
                "idx": ii,
                "label": at.A_LABEL,
                "type": at.A_SYMBOL,
                "CIP": at.A_CIP_STEREO if at.A_CIP_STEREO != "undef" else None,
                "isChiral": st,
                "isAromatic": at.A_ISAROMATIC,
                "formalCharge": at.A_FORMAL_CHARGE,
            }
            iChargeTotal += int(at.A_FORMAL_CHARGE)
            logger.debug("%d %r %r %r %r %r %r", ii, at.A_LABEL, at.A_SYMBOL, at.A_CIP_STEREO, at.A_ISAROMATIC, at.A_FORMAL_CHARGE, at.A_XYZ)
            atomL.append(aD)
        #
        bondL = []
        for ii, bnd in enumerate(eh.bonds(), 1):
            # st = bnd.B_LABEL_STEREO
            aL = bnd.atoms()
            atI = aL[0].A_LABEL
            atJ = aL[1].A_LABEL
            bD = {}
            bD = {
                "idx": ii,
                "label": bnd.B_LABEL,
                "atI": atI,
                "atJ": atJ,
                "value_order": bnd.B_ORDER,
                "CIP": bnd.B_CIP_STEREO if bnd.B_CIP_STEREO != "undef" else None,
                "isAromatic": bnd.B_ISAROMATIC,
            }
            logger.debug("%d %r %r %r %r %r %r ", ii, bnd.B_LABEL, atI, atJ, bnd.B_ORDER, bnd.B_CIP_STEREO, bnd.B_ISAROMATIC)
            bondL.append(bD)

        Prop.Setparam("E_SMILES", {"usearo": True, "usestereo": True, "unique": True})
        canIsoSmi = eh.new("E_SMILES")
        logger.debug("Canonical isomeric smiles %s", canIsoSmi)

        Prop.Setparam("E_SMILES", {"usearo": True, "usestereo": False, "unique": True})
        canSmi = eh.new("E_SMILES")
        logger.debug("Canonical smiles %s", canSmi)
        dD = {"canSmi": canSmi, "canIsoSmi": canIsoSmi}
        #
        # print(dir(eh))
        # print(eh.props())
        #
        detailD = {"ccId": eh.E_MDL_NAME, "formula": eh.E_FORMULA, "ifCharge": iChargeTotal, "formulaWeight": eh.E_WEIGHT}
        #
        retD = {"details": detailD, "descriptors": dD, "atoms": atomL, "bonds": bondL}
    except Exception as e:
        logger.exception("Failing for %s with %s", molFilePath, str(e))
    return retD


def isWritable(filePath):
    if os.path.exists(filePath):
        if os.path.isfile(filePath):
            return os.access(filePath, os.W_OK)
        else:
            return False
    dirPath = os.path.dirname(filePath)
    if not dirPath:
        dirPath = "."
    return os.access(dirPath, os.W_OK)


def main():
    if len(sys.argv) < 2:
        sys.stdout.write("script <inputMolFilePath> <outputAnnotationPath> (%r)\n" % sys.argv)
        return False
    if not os.access(sys.argv[0], os.R_OK):
        sys.stdout.write("script <inputMolFilePath> <outputAnnotationPath> (%r)\n" % sys.argv)
        return False
    #
    if isWritable(sys.argv[1]):
        retD = annotateCactvs(sys.argv[0], aroModel=sys.argv[2])
        if retD:
            with open(sys.argv[1], "w") as ofh:
                json.dump(retD, ofh)
            return True
    return False


if __name__ == "__main__":
    main()



##
# File:    ChemCompIndexProvider.py
# Author:  J. Westbrook
# Date:    16-Feb-2020
#
# Updates:
#
##
"""
Utilities to read and process an index of PDB chemical component definitions.
"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
import os
import time
from collections import defaultdict, namedtuple

from rcsb.utils.chem.ChemCompMoleculeProvider import ChemCompMoleculeProvider
from rcsb.utils.chem.OeMoleculeFactory import OeMoleculeFactory
from rcsb.utils.chem.PdbxChemComp import PdbxChemCompDescriptorIt
from rcsb.utils.chem.PdbxChemComp import PdbxChemCompIt
from rcsb.utils.chem.PdbxChemComp import PdbxChemCompAtomIt
from rcsb.utils.io.IoUtil import getObjSize
from rcsb.utils.io.MarshalUtil import MarshalUtil

# from rcsb.utils.io.SingletonClass import SingletonClass

logger = logging.getLogger(__name__)

MatchResults = namedtuple("MatchResults", "ccId oeMol searchType matchOpts screenType fpType fpScore oeIdx formula", defaults=(None,) * 9)


class ChemCompIndexProvider(object):
    """Utilities to read and process an index of PDB chemical component definitions."""

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__dirPath = os.path.join(self.__cachePath, "chem_comp")
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__ccFileNamePrefix = kwargs.get("ccFileNamePrefix", "cc")
        self.__ccIdxD = self.__reload(**kwargs)

    def getIndexFilePath(self):
        return os.path.join(self.__dirPath, "%s-idx-chemical-components.json" % self.__ccFileNamePrefix)

    def testCache(self, minCount=None, logSizes=False):
        if logSizes and self.__ccIdxD:
            logger.info("ccIdxD (%.2f MB)", getObjSize(self.__ccIdxD) / 1000000.0)
        ok = self.__ccIdxD and len(self.__ccIdxD) >= minCount if minCount else self.__ccIdxD is not None
        return ok

    def matchMolecularFormulaRange(self, typeRangeD, matchSubset=False):
        """Find matching formula for the input atom type range query (evaluates min <= ff <= max).

        Args:
            typeRangeD (dict): dictionary of element ranges {'<element_name>: {'min': <int>, 'max': <int>}}
            matchSubset (bool, optional): test for formula subset (default: False)

        Returns:
            (list):  chemical component identifiers with matching formula (MatchResults)
        """
        rL = []
        try:
            if not typeRangeD:
                return rL
            myTypeRangeD = {k.upper(): v for k, v in typeRangeD.items()}
            queryTypeS = set(myTypeRangeD.keys())
            for ccId, idxD in self.__ccIdxD.items():
                tD = idxD["type-counts"]
                targetTypeS = set(tD.keys())
                if not matchSubset and targetTypeS != queryTypeS:
                    continue
                #
                if not queryTypeS.issubset(targetTypeS):
                    continue
                #
                match = True
                for atomType, rangeD in myTypeRangeD.items():
                    if atomType in tD:
                        # min <= ff <= max
                        if ("min" in rangeD and rangeD["min"] > tD[atomType]) or ("max" in rangeD and rangeD["max"] < tD[atomType]):
                            match = False
                            break
                    else:
                        match = False
                        break
                if match:
                    # logger.info("%s formula %r query %r", ccId, idxD["type-counts"], typeRangeD)
                    rL.append(MatchResults(ccId=ccId, searchType="formula", formula=idxD["formula"]))
        except Exception as e:
            logger.exception("Failing for %r with %s", typeRangeD, str(e))
        return rL

    def filterMinimumMolecularFormula(self, typeCountD):
        """Find molecules with the minumum formula composition for the input atom type range query (evaluates min <= ff).

        Args:
            typeCountD (dict): dictionary of element minimum values {'<element_name>: #}

        Returns:
            (list):  chemical component identifiers
        """
        rL = []
        try:
            if not typeCountD:
                return list(self.__ccIdxD.keys())

            typeQueryS = set(typeCountD.keys())
            for ccId, idxD in self.__ccIdxD.items():
                tD = idxD["type-counts"]
                #
                if not typeQueryS.issubset(tD):
                    continue
                match = True
                for atomType, minCount in typeCountD.items():
                    try:
                        if minCount > tD[atomType]:
                            match = False
                            break
                    except Exception:
                        match = False
                        break
                if match:
                    rL.append(ccId)
        except Exception as e:
            logger.exception("Failing for %r with %s", typeCountD, str(e))
        return rL

    def filterMinimumFormulaAndFeatures(self, typeCountD, featureCountD):
        """Find molecules with the minumum formula and feature composition.

        Args:
            typeCountD (dict): dictionary of element minimum values {'<element_name>: #}
            featureCountD (dict): dictionary of feature minimum values {'<element_name>: #}

        Returns:
            (list):  chemical component identifiers
        """
        rL = []
        try:
            if not typeCountD or not featureCountD:
                return list(self.__ccIdxD.keys())
            # ----
            featureQueryS = set(featureCountD.keys())
            typeQueryS = set(typeCountD.keys())
            #
            for ccId, idxD in self.__ccIdxD.items():
                tD = idxD["type-counts"]
                fD = idxD["feature-counts"]
                #
                if not typeQueryS.issubset(tD) or not featureQueryS.issubset(fD):
                    continue

                match = True
                for atomType, minCount in typeCountD.items():
                    try:
                        if minCount > tD[atomType]:
                            match = False
                            break
                    except Exception:
                        match = False
                        break

                if not match:
                    continue
                #
                for featureType, minCount in featureCountD.items():
                    try:
                        if minCount > fD[featureType]:
                            match = False
                            break
                    except Exception:
                        match = False
                        break
                #
                if match:
                    rL.append(ccId)
        except Exception as e:
            logger.exception("Failing for %r with %s", typeCountD, str(e))
        return rL

    def getIndex(self):
        return self.__ccIdxD

    def getMol(self, ccId):
        try:
            return self.__ccIdxD[ccId]
        except Exception as e:
            logger.debug("Get molecule %r failing with %s", ccId, str(e))
        return None

    def __reload(self, **kwargs):
        """Reload or created index of PDB chemical components.

        Args:
            cachePath (str): path to the directory containing cache files
            ccIdxFileName (str): serialized chemical component data index file name


         Returns:
            (list): chemical component data containers
        """
        #
        logger.debug("kwargs %r", kwargs.items())
        ccIdxD = {}
        useCache = kwargs.get("useCache", True)
        molLimit = kwargs.get("molLimit", 0)

        ccIdxFilePath = self.getIndexFilePath()
        #
        if useCache and self.__mU.exists(ccIdxFilePath):
            _, fExt = os.path.splitext(ccIdxFilePath)
            ccIdxFormat = "json" if fExt == ".json" else "pickle"
            rdCcIdxD = self.__mU.doImport(ccIdxFilePath, fmt=ccIdxFormat)
            ccIdxD = {k: rdCcIdxD[k] for k in sorted(rdCcIdxD.keys())[:molLimit]} if molLimit else rdCcIdxD
        else:
            cmpKwargs = {k: v for k, v in kwargs.items() if k not in ["cachePath", "useCache", "molLimit"]}
            ccmP = ChemCompMoleculeProvider(cachePath=self.__cachePath, useCache=useCache, molLimit=molLimit, **cmpKwargs)
            ok = ccmP.testCache(minCount=molLimit, logSizes=True)
            if ok:
                molBuildType = cmpKwargs.get("molBuildType", "model-xyz")
                ccIdxD = self.__updateChemCompIndex(ccmP.getMolD(), ccIdxFilePath, molBuildType=molBuildType)
        #
        for idxD in ccIdxD.values():
            idxD["atom-types"] = set(idxD["type-counts"].keys()) if "type-counts" in idxD else set()
            idxD["feature-types"] = set(idxD["feature-counts"].keys()) if "feature-counts" in idxD else set()
        #
        return ccIdxD

    def __updateChemCompIndex(self, ccObjD, filePath, molBuildType="model-xyz"):
        idxD = {}
        try:
            # Serialized chemical component data index file
            startTime = time.time()
            _, fExt = os.path.splitext(filePath)
            fileFormat = "json" if fExt == ".json" else "pickle"
            idxD = self.__buildChemCompIndex(ccObjD, molBuildType=molBuildType)
            ok = self.__mU.doExport(filePath, idxD, fmt=fileFormat)
            endTime = time.time()
            logger.info("Storing %s with %d raw indexed definitions (status=%r) (%.4f seconds)", filePath, len(idxD), ok, endTime - startTime)
        #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return idxD

    def __buildChemCompIndex(self, cD, molBuildType="model-xyz", doFeatures=True):
        """Internal method return a dictionary of extracted chemical component descriptors and formula."""
        rD = {}
        try:
            quietFlag = True
            for _, dataContainer in cD.items():
                ccIt = iter(PdbxChemCompIt(dataContainer))
                cc = next(ccIt, None)
                ccId = cc.getId()
                formula = str(cc.getFormula()).replace(" ", "")
                ambiguousFlag = cc.getAmbiguousFlag().upper() in ["Y", "YES"]
                tch = cc.getFormalCharge()
                fcharge = int(tch) if tch and tch not in [".", "?"] else 0
                #
                logger.debug("ccId %r formula %r ambiguous %r fcharge %r", ccId, formula, ambiguousFlag, fcharge)
                if fcharge:
                    sign = "+" if fcharge > 0 else "-"
                    mag = str(abs(fcharge)) if abs(fcharge) > 1 else ""
                    formula = formula + sign + mag
                #
                atIt = PdbxChemCompAtomIt(dataContainer)
                typeCounts = defaultdict(int)
                for at in atIt:
                    aType = at.getType().upper()
                    typeCounts[aType] += 1
                #
                rD[ccId] = {"formula": formula, "type-counts": typeCounts, "ambiguous": ambiguousFlag, "feature-counts": {}}
                desIt = PdbxChemCompDescriptorIt(dataContainer)
                for des in desIt:
                    desBuildType = des.getMolBuildType()
                    tS = des.getDescriptor()
                    descr = tS.strip() if tS else None
                    if not descr:
                        continue
                    if desBuildType in ["oe-iso-smiles", "oe-smiles", "acdlabs-smiles", "cactvs-iso-smiles", "cactvs-smiles", "inchi", "inchikey"]:
                        rD[ccId][desBuildType] = descr
                    else:
                        logger.error("%s unexpected descriptor build type %r", ccId, desBuildType)
                if doFeatures:
                    oemf = OeMoleculeFactory()
                    if quietFlag:
                        oemf.setQuiet()
                    tId = oemf.setChemCompDef(dataContainer)
                    if tId != ccId:
                        logger.error("%s chemical component definition import error", ccId)
                        continue
                    ok = oemf.build(molBuildType=molBuildType)
                    if ok:
                        rD[ccId]["feature-counts"] = oemf.getFeatureCounts()

        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return rD



##
# File:    OeSubStructSearchUtils.py
# Author:  jdw
# Date:    2-Oct-2020
# Version: 0.001
#
# Updates:
#
##
"""
Utilities to manage OE specific substructure search operations (w/ formula/feature prefiltering)
"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


import logging
import time
from collections import namedtuple

from openeye import oechem

from rcsb.utils.chem.OeCommonUtils import OeCommonUtils
from rcsb.utils.chem.OeMoleculeFactory import OeMoleculeFactory
from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil

logger = logging.getLogger(__name__)

MatchResults = namedtuple("MatchResults", "ccId oeMol searchType matchOpts screenType fpType fpScore oeIdx formula", defaults=(None,) * 9)


class OeSubStructSearchWorker(object):
    """A skeleton class that implements the interface expected by the multiprocessing
    for substructure search --
    """

    def __init__(self, oeQueryMol, oeMolDb, matchOpts="graph-relaxed", **kwargs):
        self.__oeMolDb = oeMolDb
        self.__oeQueryMol = oeQueryMol
        self.__matchOpts = matchOpts
        self.__atomexpr, self.__bondexpr = OeCommonUtils.getAtomBondExprOpts(matchOpts)
        #
        _ = kwargs

    def subStructureSearch(self, dataList, procName, optionsD, workingDir):
        """Search index"""
        _ = optionsD
        _ = workingDir
        successList = []
        diagList = []
        sL = []
        # sucessList,resultList,diagList=workerFunc(runList=nextList,procName, optionsD, workingDir)
        try:
            retStatus, successList, sL = self.__subStructureSearch(self.__oeQueryMol, dataList, reverseFlag=False)
            logger.debug("%s status %r found %d search candidates from %d definitions ", procName, retStatus, len(successList), len(dataList))
        except Exception as e:
            logger.exception("Failing %s for %d data items %s", procName, len(dataList), str(e))
        #
        return successList, successList, sL, diagList

    def __subStructureSearch(self, oeQueryMol, idxList, reverseFlag=False):
        """Perform a graph match for the input query molecule on the binary
        database of molecules.  The search optionally restricted to the input index
        list.   The sense of the search may be optionally reversed.

        Args:
            oeQueryMol (object): query molecule OeGraphMol or OeQmol
            idxList ([type], optional): [description]. Defaults to None.
            reverseFlag (bool, optional): [description]. Defaults to False.
            matchOpts (str, optional): graph match criteria type (graph-strict|graph-relaxed|graph-relaxed-stereo). Defaults to "graph-relaxed".

        Returns:
            [type]: [description]
        """
        hL = []
        sL = []
        retStatus = True
        try:
            ss = oechem.OESubSearch(oeQueryMol, self.__atomexpr, self.__bondexpr)
            if not ss.IsValid():
                retStatus = False
                logger.error("Unable to initialize substructure search!")
                return retStatus, hL, sL
            #
            for idx in idxList:
                mol = oechem.OEGraphMol()
                if not self.__oeMolDb.GetMolecule(mol, idx):
                    ccId = self.__oeMolDb.GetTitle(idx)
                    logger.error("Unable to read molecule %r at index %r", ccId, idx)
                    continue
                oechem.OEPrepareSearch(mol, ss)
                if ss.SingleMatch(mol) != reverseFlag:
                    score = float(oeQueryMol.NumAtoms()) / float(mol.NumAtoms())
                    hL.append(idx)
                    sL.append(score)
            retStatus = True
        except Exception as e:
            retStatus = False
            logger.exception("Failing with %s", str(e))
        #
        return retStatus, hL, sL


class OeSubStructSearchUtils(object):
    """Utilities to manage OE specific substructure search operations (exhaustive and formula/feature prefiltered)"""

    def __init__(self, oemP, screenType=None, numProc=2, chunkSize=10, verbose=False):
        startTime = time.time()
        self.__verbose = verbose
        self.__oeMolDb, self.__oeMolDbTitleD = oemP.getOeMolDatabase()
        self.__idxTitleD = {v: k for k, v in self.__oeMolDbTitleD.items()}
        self.__numProc = numProc
        self.__chunkSize = chunkSize
        #
        if screenType:
            self.__ssDb = oemP.getSubSearchDb(screenType=screenType, numProc=numProc, forceRefresh=True)
        endTime = time.time()
        logger.info("Loaded %d definitions (%.4f seconds)", self.__oeMolDb.NumMols(), endTime - startTime)
        logger.debug("self.__oeMolDbTitleD %s", list(self.__oeMolDbTitleD.items())[:5])

    def testCache(self):
        """Check for existence of data dependencies and non-zero content counts

        Returns:
            bool: True for success or False otherwise
        """
        ok = True
        okdb = self.__oeMolDb and self.__oeMolDb.NumMols() and self.__idxTitleD and len(self.__idxTitleD)
        if not okdb:
            ok = False
        #
        logger.info("Return status %r", ok)
        return ok

    def prefilterIndex(self, oeQueryMol, idxP, matchOpts="relaxed"):
        """Filter the full search index base on minimum chemical formula an feature criteria.

        Args:
            oeQueryMol (object): search target moleculed (OEMol)
            idxP (object): instance ChemCompSearchIndexProvider()
            matchOpts (str, optional): search criteria options. Defaults to "default".

        Returns:
            (list): list of chemical component identifiers in the filtered search space
        """
        startTime = time.time()
        oemf = OeMoleculeFactory()
        oemf.setOeMol(oeQueryMol, "queryTarget")
        typeCountD = oemf.getElementCounts(useSymbol=True)
        # ccIdL1 = idxP.filterMinimumMolecularFormula(typeCountD)
        #
        featureCountD = oemf.getFeatureCounts()
        # Adjust filter according to search options
        if matchOpts in matchOpts in ["relaxed", "graph-relaxed", "simple", "sub-struct-graph-relaxed"]:
            for ky in ["rings_ar", "at_ar", "at_ch"]:
                featureCountD.pop(ky, None)
        elif matchOpts in ["relaxed-stereo", "graph-relaxed-stereo", "sub-struct-graph-relaxed-stereo"]:
            for ky in ["rings_ar", "at_ar"]:
                featureCountD.pop(ky, None)
        elif matchOpts in ["default", "strict", "graph-strict", "graph-default", "sub-struct-graph-strict"]:
            pass
        ccIdL = idxP.filterMinimumFormulaAndFeatures(typeCountD, featureCountD)
        logger.info("Pre-filtering results for formula+feature %d (%.4f seconds)", len(ccIdL), time.time() - startTime)
        return ccIdL

    def searchSubStructure(self, oeQueryMol, idxList=None, ccIdList=None, reverseFlag=False, matchOpts="graph-relaxed", numProc=1):
        if ccIdList:
            idxList = [self.__oeMolDbTitleD[ccId] for ccId in ccIdList if ccId in self.__oeMolDbTitleD]
        if numProc == 1:
            return self.__searchSubStructure(oeQueryMol, idxList=idxList, reverseFlag=reverseFlag, matchOpts=matchOpts)
        else:
            return self.__searchSubStructureMulti(oeQueryMol, idxList=idxList, matchOpts=matchOpts, numProc=numProc, maxChunkSize=self.__chunkSize)

    def __searchSubStructureMulti(self, oeQueryMol, idxList, matchOpts="graph-relaxed", numProc=2, maxChunkSize=10):
        #
        hL = []
        startTime = time.time()
        try:
            searchType = "exhaustive-substructure"
            if idxList:
                searchType = "prefilterd-substructure"
            idxList = idxList if idxList else list(range(self.__oeMolDb.GetMaxMolIdx()))
            #

            rWorker = OeSubStructSearchWorker(oeQueryMol, self.__oeMolDb, matchOpts=matchOpts)
            mpu = MultiProcUtil(verbose=True)
            optD = {"maxChunkSize": maxChunkSize}
            mpu.setOptions(optD)
            mpu.set(workerObj=rWorker, workerMethod="subStructureSearch")
            _, _, resultList, _ = mpu.runMulti(dataList=idxList, numProc=numProc, numResults=2, chunkSize=maxChunkSize)
            logger.debug("Multi-proc result length %d/%d", len(resultList[0]), len(resultList[1]))
            for idx, score in zip(resultList[0], resultList[1]):
                ccId = self.__oeMolDb.GetTitle(idx)
                hL.append(MatchResults(ccId=ccId, searchType=searchType, matchOpts=matchOpts, fpScore=score))
            retStatus = True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            retStatus = False
        logger.info("Substructure search returns %d (%.4f seconds)", len(hL), time.time() - startTime)
        return retStatus, hL

    def __searchSubStructure(self, oeQueryMol, idxList=None, reverseFlag=False, matchOpts="graph-relaxed"):
        """Perform a graph match for the input query molecule on the binary
        database of molecules.  The search optionally restricted to the input index
        list.   The sense of the search may be optionally reversed.

        Args:
            oeQueryMol (object): query molecule OeGraphMol or OeQmol
            idxList ([type], optional): [description]. Defaults to None.
            reverseFlag (bool, optional): [description]. Defaults to False.
            matchOpts (str, optional): graph match criteria type (graph-strict|graph-relaxed|graph-relaxed-stereo). Defaults to "graph-relaxed".

        Returns:
            [type]: [description]
        """
        hL = []
        retStatus = True
        startTime = time.time()
        try:
            # logger.info("Query mol type %r", type(oeQueryMol))
            atomexpr, bondexpr = OeCommonUtils.getAtomBondExprOpts(matchOpts)
            ss = oechem.OESubSearch(oeQueryMol, atomexpr, bondexpr)
            if not ss.IsValid():
                retStatus = False
                logger.error("Unable to initialize substructure search!")
                return retStatus, hL
            #
            searchType = "exhaustive-substructure"
            if idxList:
                searchType = "prefilterd-substructure"
            idxIt = idxList if idxList else range(self.__oeMolDb.GetMaxMolIdx())

            for idx in idxIt:
                mol = oechem.OEGraphMol()
                ccId = self.__oeMolDb.GetTitle(idx)
                if not self.__oeMolDb.GetMolecule(mol, idx):
                    logger.error("Unable to read molecule %r at index %r", ccId, idx)
                    continue
                oechem.OEPrepareSearch(mol, ss)
                if ss.SingleMatch(mol) != reverseFlag:
                    score = float(oeQueryMol.NumAtoms()) / float(mol.NumAtoms())
                    # hL.append(MatchResults(ccId=ccId, oeMol=mol, searchType=searchType, matchOpts=matchOpts))
                    hL.append(MatchResults(ccId=ccId, searchType=searchType, matchOpts=matchOpts, fpScore=score))
            retStatus = True
        except Exception as e:
            retStatus = False
            logger.exception("Failing with %s", str(e))
        #
        logger.info("Substructure search returns %d (%.4f seconds)", len(hL), time.time() - startTime)
        return retStatus, hL



##
# File:    OeSearchUtils.py
# Author:  jdw
# Date:    22-Oct-2019
# Version: 0.001
#
# Updates:
#
##
"""
Utilities to manage OE specific similarity search (match) operations.
"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


import logging
import time
from collections import namedtuple
from collections import OrderedDict

from openeye import oechem
from openeye import oegraphsim

from rcsb.utils.chem.OeCommonUtils import OeCommonUtils

logger = logging.getLogger(__name__)

MatchResults = namedtuple("MatchResults", "ccId oeMol searchType matchOpts screenType fpType fpScore oeIdx formula", defaults=(None,) * 9)


class OeSearchUtils(object):
    """Utilities to manage OE specific similarity search (match) operations."""

    def __init__(self, oemP, fpTypeList=None, screenType=None, numProc=2, verbose=False):
        startTime = time.time()
        self.__verbose = verbose
        self.__fpDbD = {fpType: oemP.getFingerPrintDb(fpType) for fpType in fpTypeList} if fpTypeList else {}
        self.__oeMolDb, self.__oeMolDbTitleD = oemP.getOeMolDatabase()
        self.__idxTitleD = {v: k for k, v in self.__oeMolDbTitleD.items()}
        if screenType:
            self.__ssDb = oemP.getSubSearchDb(screenType=screenType, numProc=numProc, forceRefresh=True)
        endTime = time.time()
        logger.info("Loaded %d definitions (%.4f seconds)", self.__oeMolDb.NumMols(), endTime - startTime)
        logger.debug("self.__oeMolDbTitleD %s", list(self.__oeMolDbTitleD.items())[:5])

    def testCache(self):
        """Check for existence of data dependencies and non-zero content counts

        Returns:
            bool: True for success or False otherwise
        """
        okdb = self.__oeMolDb and self.__oeMolDb.NumMols() and self.__idxTitleD and len(self.__idxTitleD)
        if not okdb or len(self.__fpDbD) < 1:
            return False
        #
        ok = True
        for fpType, fpDb in self.__fpDbD.items():
            ok = ok and fpDb.NumFingerPrints() > 1
            logger.debug("fpType %r fp count %d status %r", fpType, fpDb.NumFingerPrints(), ok)
        logger.info("Return status %r", ok)
        return ok

    def searchSubStructure(self, oeQueryMol, idxList=None, reverseFlag=False, matchOpts="graph-relaxed"):
        """Perform a graph match for the input query molecule on the binary
        database of molecules.  The search optionally restricted to the input index
        list.   The sense of the search may be optionally reversed.

        Args:
            oeQueryMol (object): query molecule OeGraphMol or OeQmol
            idxList ([type], optional): [description]. Defaults to None.
            reverseFlag (bool, optional): [description]. Defaults to False.
            matchOpts (str, optional): graph match criteria type (graph-strict|graph-relaxed|graph-relaxed-stereo). Defaults to "graph-relaxed".

        Returns:
            [type]: [description]
        """
        hL = []
        retStatus = True
        try:
            # logger.info("Query mol type %r", type(oeQueryMol))
            atomexpr, bondexpr = OeCommonUtils.getAtomBondExprOpts(matchOpts)
            ss = oechem.OESubSearch(oeQueryMol, atomexpr, bondexpr)
            if not ss.IsValid():
                retStatus = False
                logger.error("Unable to initialize substructure search!")
                return retStatus, hL
            #
            searchType = "exhaustive-substructure"
            if idxList:
                searchType = "prefilterd-substructure"
            idxIt = idxList if idxList else range(self.__oeMolDb.GetMaxMolIdx())

            for idx in idxIt:
                mol = oechem.OEGraphMol()
                ccId = self.__oeMolDb.GetTitle(idx)
                if not self.__oeMolDb.GetMolecule(mol, idx):
                    logger.error("Unable to read molecule %r at index %r", ccId, idx)
                    continue
                oechem.OEPrepareSearch(mol, ss)
                if ss.SingleMatch(mol) != reverseFlag:
                    hL.append(MatchResults(ccId=ccId, oeMol=mol, searchType=searchType, matchOpts=matchOpts))
            retStatus = True
        except Exception as e:
            retStatus = False
            logger.exception("Failing with %s", str(e))
        #
        return retStatus, hL

    def searchFingerPrints(self, oeQueryMol, fpType, minFpScore=None, maxFpResults=50, annotateMols=False, verbose=False):
        hL = []
        retStatus = True
        try:
            fpDb = self.__fpDbD[fpType] if fpType in self.__fpDbD else {}
            if not fpDb:
                retStatus = False
                return retStatus, hL
            #
            opts = oegraphsim.OEFPDatabaseOptions(maxFpResults, oegraphsim.OESimMeasure_Tanimoto)
            if minFpScore:
                opts.SetCutoff(minFpScore)
            #
            if verbose:
                logger.info("Using %d fingerprint %s type %s", fpDb.NumFingerPrints(), fpType, fpDb.GetFPTypeBase().GetFPTypeString())
                startTime = time.time()
            #
            scores = fpDb.GetSortedScores(oeQueryMol, opts)
            oeMol = oechem.OEGraphMol()
            for si in scores:
                if self.__oeMolDb.GetMolecule(oeMol, si.GetIdx()):
                    ccId = self.__oeMolDb.GetTitle(si.GetIdx())
                    if annotateMols:
                        tS = "For %s index %r %r similarity score %.4f " % (ccId, si.GetIdx(), self.__idxTitleD[si.GetIdx()], si.GetScore())
                        oechem.OESetSDData(oeMol, fpType, tS)
                    hL.append(MatchResults(ccId=ccId, oeMol=oeMol, searchType="fp", fpType=fpType, fpScore=si.GetScore()))
            if verbose:
                endTime = time.time()
                logger.info("Fingerprint %s returning %d hits (%.4f sec)", fpType, len(hL), endTime - startTime)
        except Exception as e:
            retStatus = False
            logger.exception("Failing fpType %r with %s", fpType, str(e))

        return retStatus, hL

    def getFingerPrintScores(self, oeQueryMol, fpType, minFpScore, maxFpResults):
        """Return finger print search scores for the input OE molecule.

        Args:
            oeQueryMol (OEmol): OE graph molecule
            fpType (str): fingerprint type  [TREE,PATH,MACCS,CIRCULAR,LINGO]
            fpMinScore (float): min fingerprint match score (0.0-1.0)
            maxFpResults (int): maximum number of finger print results returned

        Returns:
            (bool, list): status, finger match lists of type (MatchResults)
        """
        hL = []
        retStatus = True
        try:
            fpDb = self.__fpDbD[fpType]
            opts = oegraphsim.OEFPDatabaseOptions(maxFpResults, oegraphsim.OESimMeasure_Tanimoto)
            if minFpScore:
                opts.SetCutoff(minFpScore)
            scores = fpDb.GetSortedScores(oeQueryMol, opts)
            hL = [MatchResults(ccId=self.__oeMolDb.GetTitle(si.GetIdx()), searchType="fp", fpType=fpType, fpScore=si.GetScore(), oeIdx=si.GetIdx()) for si in scores]
        except Exception as e:
            retStatus = False
            logger.exception("Failing with %s", str(e))
        return retStatus, hL

    def searchSubStructureAndFingerPrint(self, oeQueryMol, fpTypeCutoffList, maxFpResults, matchOpts="graph-relaxed"):
        """Return graph match and finger print search results for the input OE molecule using finger print pre-filtering.

        Args:
            oeQueryMol (OEmol): OE graph molecule
            fpTypeCutoffList (list): [(finger print type, min score),...]
            maxFpResults (int): maximum number of finger print results returned
            matchOpts (str, optional): graph match criteria type (graph-strict|graph-relaxed|...). Defaults to "graph-relaxed".

        Returns:
            (bool, list, list): status, graph match and finger match lists of type (MatchResults)
        """
        fpL = []
        ssL = []
        retStatus = True
        try:
            for fpType, fpCutoff in fpTypeCutoffList:
                ok, tL = self.getFingerPrintScores(oeQueryMol, fpType, fpCutoff, maxFpResults)
                fpL.extend(tL)
            fpL = list(set(fpL))
            fpL = sorted(fpL, key=lambda nTup: nTup.fpScore, reverse=True)
            idxList = [nTup.oeIdx for nTup in fpL]
            idxList = list(OrderedDict.fromkeys(idxList))
            # -- only continue with a non-empty fingerprint result --
            if matchOpts not in ["fingerprint-similarity"] and idxList:
                # Save the maximum fp score
                fpScoreD = {}
                for fpTup in fpL:
                    fpScoreD[fpTup.ccId] = max(fpScoreD[fpTup.ccId], fpTup.fpScore) if fpTup.ccId in fpScoreD else fpTup.fpScore
                retStatus, rTupL = self.searchSubStructure(oeQueryMol, idxList=idxList, reverseFlag=False, matchOpts=matchOpts)
                for rTup in rTupL:
                    ssL.append(rTup._replace(fpScore=fpScoreD[rTup.ccId]))
        except Exception as e:
            retStatus = False
            logger.exception("Failing with %s", str(e))
        return ok and retStatus, ssL, fpL

    def searchSubStructureWithFingerPrint(self, oeQueryMol, fpType, minFpScore, maxFpResults, matchOpts="graph-relaxed"):
        """Return graph match search results for the input OE molecule using finger print pre-filtering.

        Args:
            oeQueryMol (OEmol): OE graph molecule
            fpTypeCutoffList (list): [(finger print type, min score),...]
            maxFpResults (int): maximum number of finger print results returned
            matchOpts (str, optional): graph match criteria type (graph-strict|graph-relaxed). Defaults to "graph-relaxed".

        Returns:
            (bool, list, list): status, graph match and finger match lists of type (MatchResults)
        """
        hL = []
        retStatus = True
        try:
            fpDb = self.__fpDbD[fpType]
            opts = oegraphsim.OEFPDatabaseOptions(maxFpResults, oegraphsim.OESimMeasure_Tanimoto)
            if minFpScore:
                opts.SetCutoff(minFpScore)
            scores = fpDb.GetSortedScores(oeQueryMol, opts)
            idxList = [si.GetIdx() for si in scores]
            retStatus, hL = self.searchSubStructure(oeQueryMol, idxList=idxList, reverseFlag=False, matchOpts=matchOpts)
        except Exception as e:
            retStatus = False
            logger.exception("Failing with %s", str(e))
        return retStatus, hL

    def searchSubStructureScreened(self, qmol, maxMatches=10):
        """Perform screened substructure search on input query molecule (OEQMol) subject to maxMatches.

        Args:
            qmol (OEQmol): OE query molecule
            maxMatches (int, optional): maximum number matches returned . Defaults to 10.

        Returns:
            bool, list: search return status, list of namedtuple MatchResults.
        """
        retStatus = True
        hL = []
        try:
            query = oechem.OESubSearchQuery(qmol, maxMatches)
            result = oechem.OESubSearchResult()
            status = self.__ssDb.Search(result, query)
            statusText = oechem.OESubSearchStatusToName(status)
            if statusText.upper() != "FINISHED":
                retStatus = False
                logger.info("Search failing with status = %r", statusText)
                return retStatus, hL
            #
            if self.__verbose:
                logger.info("Number of targets  = %d", result.NumTargets())
                logger.info("Number of screened = %d", result.NumScreened())
                logger.info("Number of searched = %d", result.NumSearched())
                logger.info("Number of total matches = %d", result.NumTotalMatches())
                logger.info("Number of kept  matches = %d ", result.NumMatches())
            #
            for index in result.GetMatchIndices():
                hL.append(MatchResults(ccId=self.__ssDb.GetTitle(index), searchType="screened-substructure"))
        except Exception as e:
            retStatus = False
            logger.exception("Failing with %s", str(e))

        return retStatus, hL



##
# File: PdbxChemCompConstants.py
# Date: 21-Feb-2012  John Westbrook
#
# Update:
#  21-Feb-2012 jdw add to chemcomputil repository
#   1-Feb-2017 jdw unified with chem_ref_data
#
##
"""
A collection of chemical data and information.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


class PdbxChemCompConstants(object):
    periodicTable = [
        "H",
        "HE",
        "LI",
        "BE",
        "B",
        "C",
        "N",
        "O",
        "F",
        "NE",
        "NA",
        "MG",
        "AL",
        "SI",
        "P",
        "S",
        "CL",
        "AR",
        "K",
        "CA",
        "SC",
        "TI",
        "V",
        "CR",
        "MN",
        "FE",
        "CO",
        "NI",
        "CU",
        "ZN",
        "GA",
        "GE",
        "AS",
        "SE",
        "BR",
        "KR",
        "RB",
        "SR",
        "Y",
        "ZR",
        "NB",
        "MO",
        "TC",
        "RU",
        "RH",
        "PD",
        "AG",
        "CD",
        "IN",
        "SN",
        "SB",
        "TE",
        "I",
        "XE",
        "CS",
        "BA",
        "LA",
        "CE",
        "PR",
        "ND",
        "PM",
        "SM",
        "EU",
        "GD",
        "TB",
        "DY",
        "HO",
        "ER",
        "TM",
        "YB",
        "LU",
        "HF",
        "TA",
        "W",
        "RE",
        "OS",
        "IR",
        "PT",
        "AU",
        "HG",
        "TL",
        "PB",
        "BI",
        "PO",
        "AT",
        "RN",
        "FR",
        "RA",
        "AC",
        "TH",
        "PA",
        "U",
        "NP",
        "PU",
        "AM",
        "CM",
        "BK",
        "CF",
        "ES",
        "FM",
        "MD",
        "NO",
        "LR",
        "UNQ",
        "UNP",
        "UNH",
        "UNS",
        "UNO",
        "UNE",
    ]
